/*
    libunits/units.c
    main source file

    libunits - units library
    Copyright (C) 1998-9 David A. Schleef <ds@stm.lbl.gov>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

*/

/*
	A unit is a data structure that can be used to identify a number
	with a physical unit.  One such unit is "voltage".  A unit
	consists of a baseunit and a conversion factor.  Many units
	may share the same baseunit, for example, "meter", "inch",
	"furlong", and "nanometer" are all units of length.  "Length"
	is the baseunit.  Conversion factors are the multiplier used
	to convert a unit to and from its baseunit.  Baseunits, by
	fiat, are all MKS metric units, although you do not need to
	know this and shouldn't rely on it.  To convert from
	"kilometer" to "nanometer", one would convert "kilometer" to
	the baseunit, and then convert the baseunit to "nanometer".
	Typically, you will use baseunits to communicate with other
	parts of a program, and only use units when inputting or
	outputting user data.
	
	Selection of units by name is a convenience, not a right.
	It is not intended (necessarily) to be a human-readable
	interface.  Thus, you can look up "meter", but not
	"meters" or "metre" or "m".  The reason for this is because
	1) names and plurals are locale dependent, and 2) names
	are sometimes context dependent.
*/



#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>

#define _UNITS_INTERNAL_
#include <units.h>

#define DEBUG 0

baseunit *baseunits;
baseunit baseunit0;
unit unit0;

/*
 *  doesn't do anything currently
 *  should probably read units file and/or set up standard
 *  base units
 */
void init_units(void)
{
	/* ack */
}


/*
 *  multiply units
 */
unit unit_multiply(unit a,unit b)
{
	unit u;
	
	u.to_base=a.to_base*b.to_base;
	u.from_base=a.from_base*b.from_base;
	u.base=baseunit_multiply(a.base,b.base);
	
	return u;
}

/*
 *  divide units  u=a/b
 */
unit unit_divide(unit a,unit b)
{
	unit u;
	
	u.to_base=a.to_base*b.from_base;
	u.from_base=a.from_base*b.to_base;
	u.base=baseunit_divide(a.base,b.base);
	
	return u;
}

/*
 *  multiply baseunits
 */
baseunit *baseunit_multiply(baseunit *a,baseunit *b)
{
	baseunit bu;
	int i;
	
	for(i=0;i<N_PRIM;i++){
		bu.exp[i]=a->exp[i]+b->exp[i];
	}
	
	return find_baseunit(&bu);
}

/*
 *  divide baseunits
 */
baseunit *baseunit_divide(baseunit *a,baseunit *b)
{
	baseunit bu;
	int i;
	
	for(i=0;i<N_PRIM;i++){
		bu.exp[i]=a->exp[i]-b->exp[i];
	}
	
	return find_baseunit(&bu);
}

/*
 *  returns a pointer to a persistent baseunit matching the
 *  given baseunit, possibly creating the persistent baseunit
 */
baseunit *find_baseunit(baseunit *proto)
{
	baseunit *ptr;
	
	for(ptr=baseunits;ptr;ptr=ptr->next){
		if(compare_baseunits(ptr,proto)){
			return ptr;
		}
	}
#if DEBUG
	{
		int i;
	
		printf("find_baseunit: {");
		for(i=0;i<N_PRIM;i++)printf("%d,",proto->exp[i]);
		printf("} not found\n");
	}
#endif

	/* no match, create new one */
	return new_baseunit(proto);
}

/*
 *  creates a new baseunit, linking it into list.
 *  returns NULL on error
 */
baseunit *new_baseunit(baseunit *proto)
{
	baseunit *ptr;
	
	ptr=(baseunit*)malloc(sizeof(baseunit));
	if(!ptr)return ptr;
	
	*ptr=*proto;
	ptr->next=baseunits;
	baseunits=ptr;
	
#if DEBUG
	{
		int i;
	
		printf("new_baseunit: {");
		for(i=0;i<N_PRIM;i++)printf("%d,",proto->exp[i]);
		printf("} [%p]\n",ptr);
	}
#endif
	return ptr;
}


/*
 *  compares two baseunits, returning 1 if identical, else 0
 */
int compare_baseunits(baseunit *a, baseunit *b)
{
	int i;
	
	for(i=0;i<N_PRIM;i++){
		if(a->exp[i]!=b->exp[i])return 0;
	}
	if((a->flags&UNIT_PER_ROOT_HERTZ) != (b->flags&UNIT_PER_ROOT_HERTZ))
		return 0;
	
	return 1;
}


/*
 *  string-based lookup stuff
 */

/* baseunit lookup table */

static baseunit_lookup baseunit_lookup_table[]={
	{"length",	1,  0,  0,  0,  0},
	{"mass",	0,  1,  0,  0,  0},
	{"time",	0,  0,  1,  0,  0},
	{"current",	0,  0,  0,  1,  0},
	{"temperature",	0,  0,  0,  0,  1},
	{"unitless",	0,  0,  0,  0,  0},
	{"area",	2,  0,  0,  0,  0},
	{"volume",	3,  0,  0,  0,  0},
	{"velocity",	1,  0, -1,  0,  0},
	{"acceleration",1,  0, -2,  0,  0},
	{"frequency",	0,  0, -1,  0,  0},
	{"force",	1,  1, -2,  0,  0},
	{"pressure",   -1,  1, -2,  0,  0},
	{"energy",	2,  1, -2,  0,  0},
	{"momentum",	1,  1, -1,  0,  0},
	{"power",	2,  1, -3,  0,  0},
	{"charge",	0,  0,  1,  1,  0},
	{"electric potential",	2,  1, -3, -1,  0},
	{"resistance",	2,  1, -3, -2,  0},
	{"conductance",-2, -1,  3,  2,  0},
	{"magnetic flux",		2,  1, -4, -1,  0},
	{"inductance",			2,  1, -4, -2,  0},
	{"capacitance",		       -2, -1,  4,  2,  0},
	{"electric field strength",	1,  1, -3, -1,  0},
	{"magnetic flux density",	0,  1, -4, -1,  0},
	{"electric displacement",      -2,  0,  1,  1,  0},
	{"magnetic field strength",    -1,  0,  0,  1,  0},
	{"",	0,  0,  0,  0,  0}
};
static int n_baseunit_names=
		sizeof(baseunit_lookup_table)/sizeof(baseunit_lookup)-1;

static unit_lookup unit_lookup_table[]={
	{"",		"unitless",	1.0	},
	{"meter",	"length",	1.0	},
	{"kilogram",	"mass",		1.0	},
	{"second",	"time",		1.0	},
	{"ampere",	"current",	1.0	},
	{"kelvin",	"temperature",	1.0	},
	{"newton",	"force",	1.0	},
	{"pascal",	"pressure",	1.0	},
	{"joule",	"energy",	1.0	},
	{"watt",	"power",	1.0	},
	{"coulomb",	"charge",	1.0	},
	{"volt",	"electric potential",	1.0	},
	{"ohm",		"resistance",		1.0	},
	{"weber",	"magnetic flux",	1.0	},
	{"henry",	"inductance",		1.0	},
	{"farad",	"capacitance",		1.0	},
	{"tesla",	"magnetic flux density",1.0	},
/* test units */
	{"gram",	"mass",			1e-3	},
	{"inch",	"length",		0.0254	},
	{"millimeter",	"length",		1e-3	},
	{"micron",	"length",		1e-6	},
	{"micrometer",	"length",		1e-6	},
	{"nanometer",	"length",		1e-9	},
	{"angstrom",	"length",		1e-10	},
	{"millivolt",	"electric potential",	1e-3	},
	{"microvolt",	"electric potential",	1e-6	},
	{"nanovolt",	"electric potential",	1e-9	},
	{"",		"",			0.0	}
};
static int n_unit_names=sizeof(unit_lookup_table)/sizeof(unit_lookup)-1;



/*
 *  look up unit using previous lookup table
 */
unit unit_by_name(char *name)
{
	int i;
	unit u;
	
	for(i=0;i<n_unit_names;i++){
		if(!strcmp(unit_lookup_table[i].unit_name,name)){
			u.base=baseunit_by_name(unit_lookup_table[i].baseunit_name);
			u.to_base=unit_lookup_table[i].factor;
			u.from_base=1.0/u.to_base;
			
			return u;
		}
	}
#if DEBUG
printf("unit \"%s\" not found\n",name);
#endif
	return unit0;
}

/*
 *  look up base unit by name
 */
baseunit *baseunit_by_name(char *name)
{
	int i;
	baseunit bu=baseunit0;
	
	for(i=0;i<n_baseunit_names;i++){
		if(!strcmp(baseunit_lookup_table[i].baseunit_name,name)){
			bu.exp[0]=baseunit_lookup_table[i].e0;
			bu.exp[1]=baseunit_lookup_table[i].e1;
			bu.exp[2]=baseunit_lookup_table[i].e2;
			bu.exp[3]=baseunit_lookup_table[i].e3;
			bu.exp[4]=baseunit_lookup_table[i].e4;
			
			return find_baseunit(&bu);
		}
	}
#if DEBUG
printf("baseunit \"%s\" not found\n",name);
#endif
	return NULL;
}

